@@ -9,18 +9,18 @@ from jax.lib import xla_client
 
 from ..utils import (
     HashableMPIType,
-    _constant_s32_scalar,
-    _constant_u64_scalar,
-    _ops,
-    _unpack_builder,
     default_primitive_impl,
-    dtype_ptr,
+    to_dtype_handle,
+    to_mpi_handle,
     to_mpi_ptr,
     unpack_hashable,
     wrap_as_hashable,
+    xla_constant_intc,
+    xla_constant_uintptr,
 )
+from ..decorators import translation_rule_cpu, translation_rule_gpu
 from ..validation import enforce_types
-from ..warn import warn_missing_omnistaging
+from ..comm import get_default_comm
 
 # The Jax primitive
 mpi_recv_p = Primitive("recv_mpi")  # Create the primitive
@@ -31,7 +31,7 @@ mpi_recv_impl = default_primitive_impl(mpi_recv_p)
 @enforce_types(
     source=_np.integer,
     tag=_np.integer,
-    comm=(_MPI.Intracomm, HashableMPIType),
+    comm=(type(None), _MPI.Intracomm, HashableMPIType),
     status=(type(None), _MPI.Status, HashableMPIType),
     token=(type(None), xla.Token, core.Tracer),
 )
@@ -39,15 +39,15 @@ def recv(
     x,
     source=_MPI.ANY_SOURCE,
     tag=_MPI.ANY_TAG,
-    comm=_MPI.COMM_WORLD,
+    comm=None,
     status=None,
     token=None,
 ):
-    """Perform a Recv (receive) operation.
+    """Perform a recv (receive) operation.
 
     .. warning::
 
-        Unlike mpi4py's Recv, this returns a *new* array with the received data.
+        Unlike mpi4py's recv, this returns a *new* array with the received data.
 
     Arguments:
         x: Array or scalar input with the correct shape and dtype. This can contain
@@ -55,19 +55,23 @@ def recv(
         source (int): Rank of the source MPI process.
         tag (int): Tag of this message.
         comm (mpi4py.MPI.Comm): The MPI communicator to use (defaults to
-            :obj:`COMM_WORLD`).
+            a clone of :obj:`COMM_WORLD`).
         status (mpi4py.MPI.Status): Status object, can be used for introspection.
-        token: XLA token to use to ensure correct execution order. If not given,
-            a new token is generated.
+        token (Token): XLA token to use to ensure correct execution order.
+            If not given, a new token is generated.
 
     Returns:
         Tuple[DeviceArray, Token]:
             - Received data.
             - A new, modified token, that depends on this operation.
+
     """
     if token is None:
         token = create_token(x)
 
+    if comm is None:
+        comm = get_default_comm()
+
     comm = wrap_as_hashable(comm)
 
     if status is not None:
@@ -76,44 +80,42 @@ def recv(
     return mpi_recv_p.bind(x, token, source=source, tag=tag, comm=comm, status=status)
 
 
-#  This function compiles the operation
+# This function compiles the operation
+@translation_rule_cpu
 def mpi_recv_xla_encode_cpu(c, x, token, source, tag, comm, status):
-    from ..cython.mpi_xla_bridge import MPI_STATUS_IGNORE_ADDR
-
-    warn_missing_omnistaging()
+    from ..xla_bridge.mpi_xla_bridge import MPI_STATUS_IGNORE_ADDR
 
     comm = unpack_hashable(comm)
     status = unpack_hashable(status)
 
-    c = _unpack_builder(c)
     x_shape = c.GetShape(x)
     dtype = x_shape.element_type()
     dims = x_shape.dimensions()
 
     # compute total number of elements in array
-    _nitems = _constant_s32_scalar(c, _np.prod(dims, dtype=int))
-    _dtype_ptr = dtype_ptr(dtype)
+    nitems = _np.prod(dims, dtype=int)
+    dtype_handle = to_dtype_handle(dtype)
 
     sh = xla_client.Shape.tuple_shape(
         [xla_client.Shape.array_shape(dtype, dims), xla_client.Shape.token_shape()]
     )
 
     if status is None:
-        _status = MPI_STATUS_IGNORE_ADDR
+        status_ptr = _np.uintp(MPI_STATUS_IGNORE_ADDR)
     else:
-        _status = to_mpi_ptr(status)
+        status_ptr = to_mpi_ptr(status)
 
     operands = (
-        _nitems,
-        _constant_s32_scalar(c, source),
-        _constant_s32_scalar(c, tag),
-        _constant_u64_scalar(c, to_mpi_ptr(comm)),
-        _constant_u64_scalar(c, _dtype_ptr),
-        _constant_u64_scalar(c, _status),
+        xla_constant_intc(c, nitems),
+        xla_constant_intc(c, source),
+        xla_constant_intc(c, tag),
+        xla_constant_uintptr(c, to_mpi_handle(comm)),
+        xla_constant_uintptr(c, dtype_handle),
+        xla_constant_uintptr(c, status_ptr),
         token,
     )
 
-    out = _ops.CustomCall(
+    out = xla_client.ops.CustomCall(
         c,
         b"mpi_recv",
         operands=operands,
@@ -124,43 +126,41 @@ def mpi_recv_xla_encode_cpu(c, x, token, source, tag, comm, status):
     return out
 
 
+@translation_rule_gpu
 def mpi_recv_xla_encode_gpu(c, x, token, source, tag, comm, status):
-    from ..cython.mpi_xla_bridge import MPI_STATUS_IGNORE_ADDR
-    from ..cython.mpi_xla_bridge_gpu import build_recv_descriptor
-
-    warn_missing_omnistaging()
+    from ..xla_bridge.mpi_xla_bridge import MPI_STATUS_IGNORE_ADDR
+    from ..xla_bridge.mpi_xla_bridge_gpu import build_recv_descriptor
 
     comm = unpack_hashable(comm)
     status = unpack_hashable(status)
 
-    c = _unpack_builder(c)
     x_shape = c.GetShape(x)
     dtype = x_shape.element_type()
     dims = x_shape.dimensions()
 
     # compute total number of elements in array
-    _nitems = _np.prod(dims, dtype=int)
-    _dtype_ptr = dtype_ptr(dtype)
+    nitems = _np.prod(dims, dtype=int)
+    dtype_handle = to_dtype_handle(dtype)
 
     sh = xla_client.Shape.tuple_shape(
         [xla_client.Shape.array_shape(dtype, dims), xla_client.Shape.token_shape()]
     )
 
     if status is None:
-        _status = MPI_STATUS_IGNORE_ADDR
+        status_ptr = _np.uintp(MPI_STATUS_IGNORE_ADDR)
     else:
-        _status = to_mpi_ptr(status)
+        status_ptr = to_mpi_ptr(status)
 
     descriptor = build_recv_descriptor(
-        _nitems,
+        nitems,
         source,
         tag,
-        to_mpi_ptr(comm),
-        _dtype_ptr,
-        _status,
+        to_mpi_handle(comm),
+        dtype_handle,
+        status_ptr,
     )
 
-    out = _ops.CustomCall(
+    out = xla_client.ops.CustomCall(
         c,
         b"mpi_recv",
         operands=(token,),
